import { reactive, toRefs, ref, watch, watchEffect, inject } from 'vue';
import { filterNoneValueObject, createParams, stringifyQuery, syncQuery2Conditions } from './utils';
import clone from 'rfdc';
export default function useConditionWatcher(config, queryOptions) {
    let router;
    const _conditions = reactive(config.conditions);
    const loading = ref(false);
    const data = ref(null);
    const error = ref(null);
    const refresh = ref(() => { });
    const query = ref({});
    const completeInitialConditions = ref(false);
    const fetch = (conditions) => {
        const { loading: fetchLoading, result: fetchResult, error: fetchError, use: fetchData } = useFetchData(() => config.fetcher(conditions));
        refresh.value = fetchData;
        loading.value = true;
        fetchData();
        watch([fetchResult, fetchError, fetchLoading], (values) => {
            data.value = values[0];
            error.value = values[1];
            loading.value = values[2];
        });
    };
    watchEffect(() => {
        const conditions2Object = { ..._conditions };
        let customConditions = {};
        const deepCopyCondition = clone({ proto: true })(conditions2Object);
        if (typeof config.beforeFetch === 'function') {
            customConditions = config.beforeFetch(deepCopyCondition);
            if (!customConditions || typeof customConditions !== 'object' || customConditions.constructor !== Object) {
                throw new Error(`[vue-condition-watcher]: beforeFetch should return an object`);
            }
        }
        const validateCustomConditions = Object.keys(customConditions).length !== 0;
        /*
         * if custom conditions has value, just use custom conditions
         * filterNoneValueObject will filter no value like [] , '', null, undefined
         * example. {name: '', items: [], age: 0, tags: null}
         * return result will be {age: 0}
         */
        query.value = filterNoneValueObject(validateCustomConditions ? customConditions : conditions2Object);
        const finalConditions = createParams(query.value, config.defaultParams);
        if (!completeInitialConditions.value)
            return;
        fetch(finalConditions);
    }, {
        flush: 'pre',
    });
    if (queryOptions &&
        (typeof queryOptions.sync === 'string' || typeof queryOptions.sync === 'function') &&
        queryOptions.sync.length) {
        router = inject(queryOptions.sync);
        if (router && router.isReady) {
            router.isReady().then(() => {
                // watch query changed
                watch(query, async () => {
                    const path = router.currentRoute.value.path;
                    const queryString = stringifyQuery(query.value, queryOptions.ignore || []);
                    completeInitialConditions.value = false;
                    await router.push(path + '?' + queryString);
                });
            });
            // watch router changed
            watch(router.currentRoute, (currentRoute, preCurrentRoute) => {
                const oldFullPath = preCurrentRoute ? preCurrentRoute.fullPath : '';
                if (currentRoute.fullPath !== oldFullPath) {
                    //back/forward page sync query string to _conditions
                    syncQuery2Conditions(_conditions, currentRoute.query);
                    completeInitialConditions.value = true;
                }
            }, {
                immediate: true,
            });
        }
        else {
            throw new ReferenceError('[vue-condition-watcher] Could not found vue-router instance.');
        }
    }
    completeInitialConditions.value = true;
    return {
        conditions: _conditions,
        loading,
        data,
        refresh,
        error,
    };
}
function useFetchData(fetcher) {
    const state = reactive({
        loading: false,
        error: null,
        result: null,
    });
    let lastPromise;
    const use = async () => {
        state.error = null;
        state.loading = true;
        const promise = (lastPromise = fetcher());
        try {
            const result = await promise;
            if (lastPromise === promise) {
                state.result = result;
            }
        }
        catch (e) {
            state.error = e;
        }
        finally {
            state.loading = false;
        }
    };
    return {
        ...toRefs(state),
        use,
    };
}
